/////////////////////////////////////////////////////////////////////////////
///
/// @file comp_image.cpp
///
/// An external program for computing images of cubes in the role of a
/// servant that is launched by map master for parallelized computations.
///
/// @author Ippei Obayashi, Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk and Ippei Obayashi.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 13, 2013. Last revision: April 2, 2013.


// include local header files
#include "config.h"
#ifdef CONFIG_PARALLEL
#include "utils.h"
#include "mapcomp_p.h"

// include some C++ header files
#include <iostream>
#include <sys/types.h>
#include <sys/prctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>

typedef MapServant<spcCubes, theCubMapType> Servant;
typedef chomp::multiwork::mwData mwData;


/// Suppresses all output using chomp::homology::scon, sout, sbug, etc.
/// This method is requied because stdin and stdout are used
/// to communicate with a parent process. Note that stderr (cerr) is
/// available even in this process.
void suppressOutput ()
{
	chomp::homology::sout. show = false;
	chomp::homology::scon. show = false;
	chomp::homology::serr. show = false;
	chomp::homology::slog. show = false;
	chomp::homology::sbug. show = false;
	chomp::homology::sseq. show = false;
	return;
} /* suppressOutput */

void sendOK (int fd)
{
	mwData dummy;
	SendMessage(fd, MapComm::OK, dummy);
	return;
} /* sendOK */

void sendComputedImage (int fd, mwData &outgoing_data)
{
	SendMessage(fd, MapComm::OK, outgoing_data);
	return;
} /* sendComputedImage */

void sendNG (int fd, const char* msg)
{
	mwData data;
	data << msg;
	SendMessage (fd, MapComm::NG, data);
	return;
} /* sendNG */

void computeImage (Servant &servant, mwData &incoming_data,
	mwData &outgoing_data)
{
	bool need_image = false;
	bool need_indices = false;

	incoming_data >> need_image;
	incoming_data >> need_indices;
	std::auto_ptr<spcCube> cube (MapComm::LoadCubeFrom (incoming_data));

	spcCubes image;
	std::vector<int_t> indices;
	servant.computeImage (*cube, need_image ? &image : 0,
		need_indices ? &indices : 0);

	MapComm::SaveCubeTo (outgoing_data, *cube);
	MapComm::SaveCubesTo (outgoing_data, image);
	MapComm::SaveContainerTo (outgoing_data, indices);
	return;
} /* computeImage */

void initializeMap (Servant &servant, mwData &incoming_data)
{
	int subdivDepth = 0;
	bool crop = false;
	spcCubes Xcontained, Xdisjoint;
	
	incoming_data >> subdivDepth;
	incoming_data >> crop;
	MapComm::LoadCubesFrom (incoming_data, Xcontained);
	MapComm::LoadCubesFrom (incoming_data, Xdisjoint);

	servant.initializeMap (incoming_data, subdivDepth, crop,
		Xcontained, Xdisjoint);
	return;
} /* initializeMap */

void mainLoop ()
{
	Servant servant;
	
	for (;;)
	{
		int tag;
		mwData incoming_data, outgoing_data;
		
		MapComm::RecvMessage (STDIN_FILENO, tag, incoming_data);
		
		switch (tag)
		{
		case MapComm::INITIALIZE_MAP:
			initializeMap (servant, incoming_data);
			sendOK (STDOUT_FILENO);
			break;
			
		case MapComm::COMPUTE_IMAGE:
			computeImage (servant, incoming_data, outgoing_data);
			sendComputedImage (STDOUT_FILENO, outgoing_data);
			break;
			
		case MapComm::QUIT:
			return;
		}
	}
} /* mainLoop */

void itimerHandler(int signum)
{
	if (getppid() == 1) 
		_exit(2);
}

/// Set up the process to terminate when the parent process is dead.
///
/// The life of the parent process is checked every 100ms
/// using an interval timer (SIGALRM) and getppid.
void terminateProcessIfParentIsDead ()
{
	struct sigaction sa;
	sa.sa_handler = itimerHandler;
	sa.sa_flags = SA_RESTART;
	sigemptyset(&sa.sa_mask);
	MapComm::SYSCALL ("setup SIGALRM", sigaction(SIGALRM, &sa, NULL));

	struct itimerval interval;
	interval.it_interval.tv_sec = 0;
	interval.it_interval.tv_usec = 100000; /* 100ms */
	interval.it_value = interval.it_interval;

	MapComm::SYSCALL ("setup itimer", setitimer(ITIMER_REAL, &interval, NULL));
}

int main (int argc, char *argv [])
{
	suppressOutput ();
	terminateProcessIfParentIsDead ();
        
	try
	{
		mainLoop ();
	}
	catch (const char* msg)
	{
		sendNG (STDOUT_FILENO, msg);
	//	std::cerr << "comp_image: " << msg << std::endl;
		return 1;
	}
	catch (const std::exception &e)
	{
		sendNG (STDOUT_FILENO, e.what());
	//	std::cerr << "comp_image: " << e. what() << std::endl;
		return 1;
	}
	
	return 0;
} /* main */


#else // CONFIG_PARALLEL

int main ()
{
	std::cout << "This program is in effect "
		"when CONFIG_PARALLEL is set.\n";
	return 0;
}

#endif // CONFIG_PARALLEL

