/////////////////////////////////////////////////////////////////////////////
///
/// @file mapcomp_p.h
///
/// Computation of a combinatorial map using a parallelized approach.
/// This file contains some technical functions related to spawning
/// new processes and inter-process communication.
///
/// @author Ippei Obayashi, Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk and Ippei Obayashi.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 13, 2013. Last revision: April 2, 2013.


// include local header files
#include "config.h"
#ifdef CONFIG_PARALLEL
#include "mapcomp_p.h"

// include some C++ header files
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/epoll.h>
#include <signal.h>
#include <memory>
#include <errno.h>


namespace MapComm {

/// A temporary string for a system call error message.
std::string ErrorSysCall::what_string;

/// Similar to the system call read, but this function
/// try to re-read from fd if expected amount of data
/// doesn't come.
static void Read(int fd, void* buf, size_t count)
{
	char *p = reinterpret_cast<char*>(buf);
	char *end = p + count;

	while (p < end) {
		p += SYSCALL("Read", read(fd, p, end - p));
	}
}

/// Similar to the system call write, but this function
/// try to re-write to fd if all data cannot be written.
static void Write(int fd, const void* buf, size_t count)
{
	const char *p = reinterpret_cast<const char*>(buf);
	const char *end = p + count;

	while (p < end) {
		p += SYSCALL("Write", write(fd, p, end - p));
	}
}
	
static int ReadInt(int fd)
{
	int n;
	Read(fd, reinterpret_cast<void*>(&n), sizeof(int));
	return n;
} /* ReadInt */

static void WriteInt(int fd, int n)
{
	Write(fd, reinterpret_cast<void*>(&n), sizeof(int));
} /* WriteInt */

void SendMessage(int fd, int tag, const mwData &x)
{
	WriteInt(fd, tag);
	WriteInt(fd, x.Length());
	Write(fd, x.Buffer(), x.Length());
	return;
} /* SendMessage */

void RecvMessage(int fd, int &tag, mwData &x)
{
	tag = ReadInt(fd);
	int length = ReadInt(fd);
	unsigned char* buf = new unsigned char[length];
	Read(fd, buf, length);
	x.Take(buf, length);
	return;
} /* RecvMessage */

void SaveCubeTo(mwData &data, const spcCube &cube)
{
	int dim = cube.dim();
	spcCoord c[dim];
	cube.coord(c);
	data << dim;
	for (int i=0; i<dim; ++i) 
		data << c[i];
	return;
} /* SaveCubeTo */

spcCube* LoadCubeFrom(mwData &data)
{
	spcCube *cube = new spcCube();
	LoadCubeFrom(data, *cube);
	return cube;
} /* LoadCubeFrom */

void LoadCubeFrom(mwData &data, spcCube &cube)
{
	int dim = 0;
	data >> dim;

	spcCoord c[dim];
	for (int i=0; i<dim; ++i)
		data >> c[i];
	cube = spcCube(c, dim);
	return;
} /* LoadCubeFrom */

void SaveCubesTo(mwData &data, const spcCubes & cubes)
{
	data << cubes.size();
	for (int_t cur=0; cur < cubes.size(); ++cur) {
		SaveCubeTo(data, cubes[cur]);
	}
	return;
} /* SaveCubesTo */

spcCubes* LoadCubesFrom(mwData &data)
{
	spcCubes* cubes = new spcCubes();
	LoadCubesFrom(data, *cubes);
	return cubes;
} /* LoadCubesFrom */

void LoadCubesFrom(mwData &data, spcCubes &cubes)
{
	int_t size = 0;
	data >> size;

	for (int_t cur=0; cur < size; ++cur) {
		std::auto_ptr<spcCube> cube(LoadCubeFrom(data));
		cubes.add(*cube);
	}
	return;
} /* LoadCubesFrom */

static void startServantProcess(int readpipe[], int writepipe[],
			      const char* path,
			      const char * const* argv)
{
	close(readpipe[0]);
	close(writepipe[1]);
	dup2(readpipe[1], STDOUT_FILENO);
	dup2(writepipe[0], STDIN_FILENO);
	
	SYSCALL("execvp", execvp(path, const_cast<char * const*>(argv)));
	return;
} /* startServantProcess */

ServantProcess* spawnProcess(const char* path, const char * const* argv, int id)
{
	pid_t pid;
	int readpipe[2], writepipe[2];

	SYSCALL("read pipe", pipe(readpipe));
	SYSCALL("write pipe", pipe(writepipe));

	pid = SYSCALL("fork", fork());
	
	if (pid == 0) {
		startServantProcess(readpipe, writepipe, path, argv);
		_exit(9);
	}
	
	close(readpipe[1]);
	close(writepipe[0]);

	ServantProcess *process = new ServantProcess();
	process->id = id;
	process->readfd = readpipe[0];
	process->writefd = writepipe[1];
	process->pid = pid;
	return process;
} /* spawnProcess */

void shutdownProcess(ServantProcess* process, bool send_signal)
{
	int status;
	if (send_signal) {
		kill(process->pid, SIGTERM);
	} else {
		mwData empty_data;
		SendMessage(process->writefd, MapComm::QUIT, empty_data);
	}
	close(process->readfd);
	close(process->writefd);

	waitpid(process->pid, &status, 0);
	delete process;
	return;
} /* shutdownProcess */

void decodeError(mwData &data)
{
	static char buf[16384];
	data >> buf;
	throw buf;
} /* decodeError */


} // namespace MapComm

#endif // CONFIG_PARALLEL

