/////////////////////////////////////////////////////////////////////////////
///
/// @file mapcomp_p.h
///
/// Computation of a combinatorial map using a parallelized approach.
/// This file contains definitions of classes that can be used
/// to compute images of cubes in a distributed way.
///
/// @author Ippei Obayashi, Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk and Ippei Obayashi.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 13, 2013. Last revision: March 24, 2013.


#ifndef _CMGRAPHS_MAPCOMP_P_H_
#define _CMGRAPHS_MAPCOMP_P_H_

#ifdef CONFIG_PARALLEL

// include some standard C++ header files
#include <memory>
#include <cassert>
#include <algorithm>
#include <functional>
#include <poll.h>
#include <exception>
#include <string>

// include selected header files from the CHomP library
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/cubes/cube.h"
#include "chomp/multiwork/mwdata.h"
#include "chomp/struct/digraph.h"

// include local header files
#include "config.h"
//#include "typedefs.h"
//#include "typedyns.h"
//#include "utils.h"
#include "datavector.h"


namespace MapComm {

/// An exception that is thrown in case of a system call error.
class ErrorSysCall: public std::exception
{
public:
	/// The constructor.
	ErrorSysCall (const std::string &_details) throw ():
		details (_details) {};

	/// The destructor.
	~ErrorSysCall () throw () {};

	/// Description of the error.
	const char *what() const throw ()
	{
		what_string = std::string ("System call failure: ") +
			details + ".";
		return what_string. c_str ();
	};

private:
	/// Details of the system call error.
	std::string details;

	/// A static string for returning the text message.
	static std::string what_string;

}; /* ErrorSysCall */

enum ServiceType { INITIALIZE_MAP, COMPUTE_IMAGE, QUIT };

enum Reply { OK, NG };

typedef chomp::multiwork::mwData mwData;

inline int SYSCALL(const char *details, int retval)
{
	if (retval < 0) {
		perror("");
		throw ErrorSysCall (details);
	//	throw "system call error";
	}
	return retval;
} /* SYSCALL */

/// Sends a message to fd (file descriptor of unix system).
/// You can reproduce the message using RecvMessage.
void SendMessage(int fd, int tag, const mwData &x);

/// Receives a message from fd sent by SendMessage.
void RecvMessage(int fd, int &tag, mwData &x);

/// Serializes spcCubes to mwData.
void SaveCubesTo(mwData &x, const spcCubes &cubes);

/// Serializes spcCubes to mwData.
void SaveCubeTo(mwData &data, const spcCube &cube);
    
/// Deserializes spcCubes from mwData.
spcCubes* LoadCubesFrom(mwData &data);

/// Deserializes spcCubes from mwData.
void LoadCubesFrom(mwData &data, spcCubes &cubes);

/// Deserializes spcCube from mwData.
spcCube* LoadCubeFrom(mwData &data);

/// Deserializes spcCube from mwData.
void LoadCubeFrom(mwData &data, spcCube &cube);

/// Serializes any types of c++ containers to mwData.
/// The value type of the container should be serializable by mwData.
template<class container>
void SaveContainerTo(mwData &data, const container &v)
{
	typedef typename container::const_iterator iter;

	data << static_cast<int> (v.size());
	for (iter i = v.begin(); i != v.end(); ++i) {
		data << *i;
	}
	return;
} /* SaveContainerTo */

/// Deserializes any types of c++ containers with push_back from mwData.
/// The value type of the container should be serializable by mwData.
template<class container>
void LoadContainerFrom(mwData &data, container & v)
{
	int size = 0;
	data >> size;
	for (int i=0; i<size; i++) {
		typename container::value_type value = 0;
		data >> value;
		v.push_back(value);
	}
} /* LoadContainerFrom */

struct ServantProcess
{
	int id;
	int readfd;
	int writefd;
	pid_t pid;

}; /* struct ServantProcess */

/// Spawns a new process, prepares input and output pipes, and
/// returns a new ServantProcess object.
/// The return value should be freed by shutdownProcess.
ServantProcess* spawnProcess(const char* path, const char *const* argv, int id);

/// Shuts down the given subprocess and frees the memory
/// of the process data structure.
/// If use_signal is true, the process is forced to be terminated
/// with SIGTERM signal. If use_signal is false, the process
/// is gracefully terminated.
void shutdownProcess(ServantProcess* process, bool use_signal);

/// Decodes an error message from mwData and throws the message.
void decodeError(mwData &data);

/// The number of parallel computations
#if CONFIG_PARALLEL > 0
static const int numParallels = CONFIG_PARALLEL;
#else
static const int numParallels = 4
#endif
		
} /* namespace MapComm */


// --------------------------------------------------
// ------------------ map servant -------------------
// --------------------------------------------------

/// A class whose objects may work as servers for parallelized computation
/// of the graph of a combinatorial cubical multivalued map.
template <class typeCubes, class typeCubMap>
class MapServant
{
public:
	/// The default constructor.
	MapServant ();

	/// The destructor.
	~MapServant ();

	/// Initializes the object for parallelized map computation.
	/// Receives information on how to create the correct map,
	/// including the subdivision depth and the information
	/// on whether the images should be cropped to the phase space.
	/// Additionally two sets of cubes may be provided (they are only
	/// in effect if nonempty): one is a set in which all the images
	/// should be contained, another is a set with which
	/// the images should be disjoint.
	/// In case of doing computations locally, the initialization
	/// may be different (e.g. without affecting some global variables).
	void initializeMap (chomp::multiwork::mwData &_mapData,
			    int _subdivDepth, bool _crop,
			    typeCubes &_Xcontained, typeCubes &_Xdisjoint);

	/// Computes the image of a single cube.
	/// Adds the cubes in the image to the set of cubes
	/// (unless the pointer to the set of cubes is null).
	/// If 'Xcontained' is nonempty then fills in the vector
	/// of indices of cubes in the map (unless the pointer is null).
	/// Throws an exception in case of failure.
	void computeImage (const typename typeCubes::value_type &q,
			   typeCubes *img, std::vector<int_t> *indices) const;
        
        
private:
	/// The copy constructor should not be used.
	MapServant (const MapServant<typeCubes,typeCubMap> &);

	/// The assignment operator should not be used.
	MapServant<typeCubes,typeCubMap> &operator =
	(const MapServant<typeCubes,typeCubMap> &);

	/// A set of cubes in which the computed images should be contained.
	/// If this set is empty then this condition is not in effect.
	typeCubes Xcontained;

	/// A set of cubes with which the computed images should be disjoint.
	/// If this set is empty then this condition is not in effect.
	typeCubes Xdisjoint;

	/// The phase space offset as an array.
	double offset [spaceDim];

	/// The phase space width as an array.
	double width [spaceDim];

	/// The interval map whose internals must be set each time.
	typename typeCubMap::MapCompType theMap;

	/// The cubical map that is based upon the interval map.
	typeCubMap *theCubMap;

}; /* class MapServant */

// --------------------------------------------------

template <class typeCubes, class typeCubMap>
inline MapServant<typeCubes,typeCubMap>::MapServant ():
		theCubMap (0)
{
	// prepare the offset and the width of the phase space region
	for (int i = 0; i < spaceDim; ++ i)
	{
		offset [i] = spaceOffset [i];
		width [i] = spaceWidth [i];
	}

	return;
} /* MapServant::MapServant */

template <class typeCubes, class typeCubMap>
inline MapServant<typeCubes,typeCubMap>::~MapServant ()
{
	if (theCubMap)
		delete theCubMap;
	return;
} /* MapServant::~MapServant */


template <class typeCubes, class typeCubMap>
inline void MapServant<typeCubes,typeCubMap>::initializeMap
(chomp::multiwork::mwData &_mapData, int _subdivDepth, bool _crop,
 typeCubes &_Xcontained, typeCubes &_Xdisjoint)
{
	// set the internal parameters for the interval map
	theMap. loadInternals (_mapData);

	// take the provided sets of cubes (note: avoid duplication of data!)
	Xcontained. swap (_Xcontained);
	Xdisjoint. swap (_Xdisjoint);

	// create the combinatorial cubical multivalued map object
	if (theCubMap)
		delete theCubMap;
	theCubMap = new typeCubMap (offset, width,
				    1 << _subdivDepth, _subdivDepth, theMap);
	theCubMap -> cache = false;

	// set space wrapping in the servant (note: this is a global setting
	// that may affect the main program, unless the servant is in fact
	// used in a separate process)
	setSpaceWrapping (theCubMap -> intWidth ());

	// set map cropping in order to avoid processing the image cubes
	// outside the phase space (if requested)
	theCubMap -> cropping = _crop;

	return;
} /* MapServant::initializeMap */

template <class typeCubes, class typeCubMap>
inline void MapServant<typeCubes,typeCubMap>::computeImage
(const typename typeCubes::value_type &q, typeCubes *img,
 std::vector<int_t> *indices) const
{
	// throw an exception if requested to compute the indices
	// but not enough information for that is provided
	if (indices && Xcontained. empty ())
		throw "Can't compute indices of cubes in the empty set.";

	// create a fake graph with one vertex, just to capture the image
	chomp::homology::diGraph<> g;
	if (indices)
		g. addVertex ();

	// compute the image of the cube
	(*theCubMap) (q, img, indices ? &g : 0,
			!Xcontained. empty () ? &Xcontained : 0,
			!Xdisjoint. empty () ? &Xdisjoint : 0, false);

	// put indices of the computed cubes in the vector
	if (indices)
	{
		int_t nEdges = g. countEdges (0);
		indices -> resize (nEdges);
		for (int_t edge = 0; edge < nEdges; ++ edge)
			(*indices) [edge] = g. getEdge (0, edge);
	}

	return;
} /* MapServant::computeImage */


// --------------------------------------------------
// ------------------- map master -------------------
// --------------------------------------------------

/// A class whose object works as a master for parallelized computation
/// of the graph of a combinatorial cubical multivalued map.
template <class typeCubes, class typeCubMap>
class MapMaster
{
public:
	typedef typename typeCubes::value_type typeCube;
	
	/// The default constructor.
	MapMaster (const int _nParallel,
		   const typeCubMap &cubMap, bool _crop,
		   const typeCubes &_Xcontained, const typeCubes &_Xdisjoint,
		   bool _need_image, bool _need_indices);

	/// The destructor.
	~MapMaster ();

	/// Queues the qNew to the queue of cubes to be computed
	/// their images. If there is a free servant, the work is
	/// assigned to the servant. If no servant is free,
	/// the work is only queued (reserved).
	/// In the both cases, the function doesn't wait the
	/// servants to finish their work and queueCubeToCompute
	/// returns immediately.
	void requestComputation(const typeCube &qNew);

	/// Waits for a servant to finish his work to compute the image,
	/// sets the result to qComputed, imgComputed, and indices,
	/// and returns true.
	/// imgComputed are filled if need_image is true.
	/// indices are filled if need_indices is true.
	/// If the images of all cubes are computed, returns false.
	bool waitResult(typeCube &qComputed,
			typeCubes* imgComputed, std::vector<int_t>* indices);

	/// Returns true if all works are finished.
	bool finished(){ return queue.empty() && working_servants.empty(); }
	
	/// Stops the parallelized computations. In particular,
	/// releases the servants, closes network connections, etc.
	void stop ();

private:
	/// The copy constructor should not be used.
	MapMaster (const MapMaster<typeCubes,typeCubMap> &);

	/// The assignment operator should not be used.
	MapMaster<typeCubes,typeCubMap> &operator =
	(const MapMaster<typeCubes,typeCubMap> &);

	/// Spawn servant processes and add them to servants
	/// and free_servants.
	void spawnServantProcesses();

	/// Initialize all servants.
	void initializeServantProcesses();

	/// Initialize one servant.
	void initializeServantProcess(MapComm::ServantProcess *servant);
	
	/// Examine whether there is a servant process who finish
	/// his assigned work. If so, the result ot the process
	/// are set to qComputed, imgComputed, indices, and
	/// true is returned. Otherwise, false is returned.
	bool peekWorkingProcesses(typeCube &qComputed,
				  typeCubes *imgComputed,
				  std::vector<int_t> *indices)
	{
		return pollWorkingProcesses(qComputed, imgComputed, indices, 0);
	}

	/// Wait for a servant process to finish his work,
	/// the result is assigned to given parameters,
	/// and return true.
	/// If there is no working process, return false.
	bool waitWorkingProcesses(typeCube &qComputed,
				  typeCubes *imgComputed,
				  std::vector<int_t> *indices)
	{
		return pollWorkingProcesses(qComputed, imgComputed, indices, -1);
	}

	bool pollWorkingProcesses(typeCube &qComputed,
				  typeCubes *imgComputed,
				  std::vector<int_t> *indices,
				  int timeout);

	void readResultFromServant(MapComm::ServantProcess* servant,
				   typeCube &qComputed,
				   typeCubes &imgComputed,
				   std::vector<int_t> &indices)
	{
		chomp::multiwork::mwData data;
		int tag;
		
		MapComm::RecvMessage(servant->readfd, tag, data);

		if (tag == MapComm::NG)
			MapComm::decodeError(data);
		
		MapComm::LoadCubeFrom(data, qComputed);
		MapComm::LoadCubesFrom(data, imgComputed);
		MapComm::LoadContainerFrom(data, indices);
	}

	/// Assigns a work (identified by q) to a free servant process.
	/// You should verify that there is a free servant process
	/// before calling this method.
	void assignWorkToServant(const typeCube &q);
	
	/// A suggested number of parallel jobs to do.
	const int nParallel;

	/// The information about internals of the map to use.
	chomp::multiwork::mwData mapData;

	/// The current subdivision level.
	const int subdivDepth;

	/// Should the images be cropped to the phase space?
	const bool crop;

	/// A set of cubes in which the computed images should be contained.
	/// If this set is empty then this condition is not in effect.
	typeCubes Xcontained;

	/// A set of cubes with which the computed images should be disjoint.
	/// If this set is empty then this condition is not in effect.
	typeCubes Xdisjoint;

	/// A switch to determine that the image of a cube (imgComputed)
	/// is required.
	bool need_image;

	/// A switch to determine that the indices of cubes in the image
	/// of a cube are required.
	bool need_indices;
	
	/// A set of servant processes
	std::vector<MapComm::ServantProcess*> servants;

	/// A set of free servant processes.
	std::vector<MapComm::ServantProcess*> free_servants;

	/// A set of working servant processes.
	std::vector<MapComm::ServantProcess*> working_servants;

	/// The queue of cubes to be handled.
	std::queue<typeCube> queue;

}; /* class MapMaster */

// --------------------------------------------------

template <class typeCubes, class typeCubMap>
inline MapMaster<typeCubes,typeCubMap>::MapMaster
(const int _nParallel,
 const typeCubMap & cubMap,
 bool _crop,
 const typeCubes &_Xcontained,
 const typeCubes &_Xdisjoint,
 bool _need_image, bool _need_indices):
		nParallel (_nParallel),
		mapData (),
		subdivDepth (cubMap.subdivDepth()),
		crop (_crop),
		Xcontained (_Xcontained),
		Xdisjoint (_Xdisjoint),
		need_image(_need_image),
		need_indices(_need_indices)
{
	cubMap.M.saveInternals(mapData);
	return;
} /* MapMaster::MapMaster */

template <class typeCubes, class typeCubMap>
inline MapMaster<typeCubes,typeCubMap>::~MapMaster ()
{
	// release any active servants
	stop ();

	return;
} /* MapMaster::~MapMaster */

template <class typeCubes, class typeCubMap>
void MapMaster<typeCubes,typeCubMap>::
requestComputation(const typename typeCubes::value_type &qNew)
{
	if (servants.empty()) {
		spawnServantProcesses();
		initializeServantProcesses();
	}
	
	if (free_servants.empty()) {
		queue.push(qNew);
	} else {
		assignWorkToServant(qNew);
	}
}

template <class typeCubes, class typeCubMap>
bool MapMaster<typeCubes,typeCubMap>::
waitResult(typename typeCubes::value_type &qComputed,
	   typeCubes* imgComputed, std::vector<int_t>* indices)
{
	if (working_servants.empty())
		return false;

	waitWorkingProcesses(qComputed, imgComputed, indices);

	if (!queue.empty()) {
		typeCube qNew = queue.front();
		queue.pop();
		assignWorkToServant(qNew);
	}

	return true;
}

template <class typeCubes, class typeCubMap>
inline void MapMaster<typeCubes,typeCubMap>::stop ()
{
	std::for_each(servants.begin(), servants.end(),
		      std::bind2nd(std::ptr_fun(MapComm::shutdownProcess),
				   !working_servants.empty()));
	servants.clear();
} 


template <class typeCubes, class typeCubMap>
void MapMaster<typeCubes,typeCubMap>::spawnServantProcesses()
{
	// check if the servant program exists
	{
		std::ifstream in ("./comp_image");
		if (!in)
		{
			throw MapComm::ErrorSysCall
				("'./comp_image' not found");
		}
	}

	static const char * path = "./comp_image";
	static const char * const argv[] = {
		"./comp_image", NULL,
	};
	
	for (int i=0; i<nParallel; ++i) {
		servants.push_back(MapComm::spawnProcess(path, argv, i));
	}
	std::copy(servants.begin(), servants.end(),
		  std::back_inserter(free_servants));
}

template <class typeCubes, class typeCubMap>
void MapMaster<typeCubes,typeCubMap>::initializeServantProcess(
    MapComm::ServantProcess *servant)
{
	chomp::multiwork::mwData outgoing_data, incoming_data;
	outgoing_data << subdivDepth;
	outgoing_data << crop;
	MapComm::SaveCubesTo(outgoing_data, Xcontained);
	MapComm::SaveCubesTo(outgoing_data, Xdisjoint);
	outgoing_data << mapData;
	MapComm::SendMessage(servant->writefd, MapComm::INITIALIZE_MAP,
			     outgoing_data);
	int tag;
	MapComm::RecvMessage(servant->readfd, tag, incoming_data);

	if (tag == MapComm::NG)
		MapComm::decodeError(incoming_data);
}

template <class typeCubes, class typeCubMap>
void MapMaster<typeCubes,typeCubMap>::initializeServantProcesses()
{
	typedef std::vector<MapComm::ServantProcess*>::iterator iter;
	for (iter p = servants.begin(); p != servants.end(); ++p)
		initializeServantProcess(*p);
}

template <class typeCubes, class typeCubMap>
bool MapMaster<typeCubes,typeCubMap>::pollWorkingProcesses
(typename typeCubes::value_type &qComputed,
 typeCubes *imgComputed, std::vector<int_t> *indices,
 int timeout)
{
	using MapComm::SYSCALL;
	
	struct pollfd fds[nParallel];
	for (unsigned int i=0; i<working_servants.size(); ++i) {
		memset(&fds[i], 0, sizeof(struct pollfd));
		fds[i].fd = working_servants[i]->readfd;
		fds[i].events = POLLIN;
	}

	int num_event = SYSCALL("poll", poll(fds, working_servants.size(), timeout));

	if (num_event == 0)
		return false;

	typeCubes dummy_cubes;
	std::vector<int_t> dummy_indices;
	imgComputed = imgComputed ? imgComputed : &dummy_cubes;
	indices = indices ? indices : &dummy_indices;

	for (unsigned int i=0; i<working_servants.size(); ++i) {
		if (fds[i].revents & POLLIN) {
			readResultFromServant(working_servants[i],
					      qComputed, *imgComputed, *indices);
			free_servants.push_back(working_servants[i]);
			working_servants.erase(working_servants.begin() + i);
			return true;
		}
	}
	
	return false;
}

template <class typeCubes, class typeCubMap>
void MapMaster<typeCubes,typeCubMap>::assignWorkToServant
(const typename typeCubes::value_type &q)
{
	assert(!free_servants.empty());

	chomp::multiwork::mwData outgoing_data;
	outgoing_data << need_image;
	outgoing_data << need_indices;
	MapComm::SaveCubeTo(outgoing_data, q);

	MapComm::ServantProcess* servant = free_servants.back();
	free_servants.pop_back();
	
	MapComm::SendMessage(servant->writefd, MapComm::COMPUTE_IMAGE,
			     outgoing_data);

	working_servants.push_back(servant);
}

#endif // CONFIG_PARALLEL

#endif // _CMGRAPHS_MAPCOMP_P_H_

// Local Variables:
// mode: c++
// indent-tabs-mode: t
// c-basic-offset: 8
// c-file-offsets: ((innamespace . 0) (inclass . +))
// End:
