/////////////////////////////////////////////////////////////////////////////
///
/// @file mapgraph_p.h
///
/// Computation of the graph representation of a combinatorial map
/// using a parallelized approach.
/// This file contains the definition of a function for the computation
/// of a directed graph which represents a combinatorial map
/// restricted to a given set of cubes.
///
/// @author Ippei Obayashi, Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk and Ippei Obayashi.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on February 13, 2013. Last revision: March 17, 2013.


#ifndef _CMGRAPHS_MAPGRAPH_P_H_
#define _CMGRAPHS_MAPGRAPH_P_H_


// include some standard C++ header files
#include <memory>
#include <map>
#include <deque>

// include selected header files from the CHomP library
#include "chomp/system/textfile.h"
#include "chomp/cubes/pointset.h"
#include "chomp/cubes/cube.h"
#include "chomp/multiwork/mwdata.h"

// include local header files
#include "config.h"
#include "typedefs.h"
#include "typedyns.h"
#include "utils.h"
#include "spacewrap.h"
#include "datavector.h"
#include "mapcomp_p.h"


// --------------------------------------------------
// -------- compute cubical map in parallel ---------
// --------------------------------------------------

/// Computes the cubical map on 'X' and fills out the graph 'g'
/// using parallelized computing as much as possible.
template <class typeCubes, class typeGraph, class typeCubMap>
inline void computeMapGraphParallel (const typeCubes &X,
	typeGraph &g, const typeCubMap &theCubMap)
{
	typedef typename typeCubes::value_type typeCube;
	
	// create a map master that will distribute the computations
	typeCubes empty;
	MapMaster<typeCubes,typeCubMap> mapMaster(
	    MapComm::numParallels, theCubMap, true, X, empty, false, true);

	// prepare a data structure based on standard C++ containers
	// that will temporarily keep the computed graph
	std::deque<std::vector<int_t> > results(X.size());

	// give all the cubes for the master so that it does the computations
        for (int_t cur = 0; cur < X.size(); ++cur) 
		mapMaster.requestComputation(X[cur]);

	// compute the multivalued map and store it
	// in the temporary data structure
	while (!mapMaster.finished()) {
		typeCube qComputed;
		std::vector<int_t> indices;
		if (mapMaster.waitResult(qComputed, 0, &indices)) {
			results[X.getnumber(qComputed)] = indices;
		}
	}

	// copy the temporary data structure to the graph of the map
	for (int_t cur = 0; cur < X.size(); ++cur) {
		g.addVertex();
		const std::vector<int_t> & result = results.front();
		
		typedef std::vector<int_t>::const_iterator iter;
		for (iter p = result.begin(); p != result.end(); ++p) {
			g.addEdge(*p);
		}
		results.pop_front();
	}

	// stop the processes started for the parallelized computations
	mapMaster. stop ();

	return;
} /* computeMapGraph */


#endif // _CMGRAPHS_MAPGRAPH_P_H_

// Local Variables:
// mode: c++
// indent-tabs-mode: t
// c-basic-offset: 8
// c-file-offsets: ((innamespace . 0) (inclass . +))
// End:


