/////////////////////////////////////////////////////////////////////////////
///
/// @file p_inf.h
///
/// Parameters for an infectious disease ODE.
///
/// @author Pawel Pilarczyk
///
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2014 by Pawel Pilarczyk.
//
// This file is part of my research software package.  This is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This software is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on March 13, 2013. Last revision: July 8, 2014.


#ifndef _CMGRAPHS_P_H_
#define _CMGRAPHS_P_H_


namespace custom {

/// Parameters for the two coupled quintic Vanderpol maps.
namespace InfModel {


// --------------------------------------------------
// --------- the phase space configuration ----------
// --------------------------------------------------

/// The dimension of the phase space.
const int spaceDim = 4;

/// Is space wrapping (a.k.a. periodic boundary conditions)
/// in effect in the given directions?
const bool spaceWrapping [spaceDim] = {false, false, false, false};

/// The initial depth of subdivisions in the phase space.
const int initialDepth = 3;

/// The final depth of subdivisions in the phase space.
const int finalDepth = 9;


// --------------------------------------------------
// ------------ the parameter selection -------------
// --------------------------------------------------

/// The dimension of the parameter space to iterate. Only those parameters
/// which are varying are taken into consideration here.
const int paramDim = 2;

/// The numbers of subintervals in each direction of the parameter space.
/// The type of these numbers must match "parCoord" in "typedefs.h".
const short int paramSubdiv [paramDim] = {200, 200};

/// The numbers of parameters to subdivide.
const int paramSelect [paramDim] = {2, 4};


// --------------------------------------------------
// ---------- the parameter space settings ----------
// --------------------------------------------------

/// The number of all the parameters, both varying and fixed.
// Following (7) from the notes from Oct 13, 2013.
const int paramCount = 16;
// K1, K2, alpha1, alpha2, beta1, beta2,
// mu1, mu2, gamma1, gamma2, theta1, theta2, sigma1, sigma2, phi1, phi2;
// note: alpha1*k1=alpha2*k2 for the 4D model to be valid

/// The left bounds on the parameters.
//const double paramLeft [paramCount] = {100, 100, 1.0/7, -1, 0.5, -1,
//	0.1, 0.1, 12, 12, 0.5, 0.5, 0.2, 0.2, 16, 16};
const double paramLeft [paramCount] = {100, 100, 0, -1, 0.48, -1,
	0.1, 0.1, 12, 12, 0.5, 0.5, 0.2, 0.2, 16, 16};

/// The right bounds on the parameters.
//const double paramRight [paramCount] = {100, 100, 1.0/7, -1, 0.5, -1,
//	0.1, 0.1, 12, 12, 0.5, 0.5, 0.2, 0.2, 16, 16};
const double paramRight [paramCount] = {100, 100, 0.5, -1, 0.53, -1,
	0.1, 0.1, 12, 12, 0.5, 0.5, 0.2, 0.2, 16, 16};

#define ALPHA1_EQ_ALPHA2
#define BETA1_EQ_BETA2

/* #define NONLINEAR_PARAMETERS
inline double nonlinearParameter (int j, int c)
{
	int i = paramSelect [j];
	double paramWidth = paramRight [i] - paramLeft [i];
	if (i == 2)
	{
		return (paramLeft [i] + (c * paramWidth * c /
			paramSubdiv [j]) / paramSubdiv [j]);
	}
	else
	{
		return (paramLeft [i] + c * paramWidth / paramSubdiv [j]);
	}
}*/ /* nonlinearParameter */


// --------------------------------------------------
// -------------- the phase space size --------------
// --------------------------------------------------

/// The offset of the rectangular area
/// in the phase space which contains the invariant sets of interest.
const double spaceOffset [spaceDim] = {0, 0, 0, 0};

/// The width of the rectangular area
/// in the phase space which contains the invariant sets of interest.
const double spaceWidth [spaceDim] = {100, 100, 100, 100};


// --------------------------------------------------
// ------------------ refinements -------------------
// --------------------------------------------------

/// The number of refinements that should be done if a Morse set
/// with the trivial index is encountered or if the image of a Morse set
/// sticks out of the rectangular region to which the computations
/// are restricted. Set to zero for no refinements.
const int refineDepth = 5;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the initial subdivision level.
/// Set to zero to suppress any refinements.
const int maxRefineSize0 = 3000;

/// The maximal allowed size of a set of cubes in the phase space
/// which can be refined at the subsequent subdivision levels.
/// Set to zero to suppress any refinements.
const int maxRefineSize1 = 3000;


// --------------------------------------------------
// ------------------- map images -------------------
// --------------------------------------------------

/// The maximal allowed diameter of the cubical image of a signle box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
/// Should be at least 3 to accept repelling boxes.
const int maxImageDiameter = 50;

/// The maximal allowed volume of the cubical image of a single box.
/// This limit is set up in order to prevent a blow-up of the program
/// in the memory because of accumulating huge images of boxes.
/// Should be at least 3^spaceDim to accept repelling boxes.
const int maxImageVolume = 361;


// --------------------------------------------------
// -------------- max index pair size ---------------
// --------------------------------------------------

/// The maximal allowed size of the index pair. This size is measured
/// as the total number of cubes in the index pair and its forward image.
const int maxIndexPairSize = 10000;


// --------------------------------------------------
// --------------- joining Morse sets ---------------
// --------------------------------------------------

/// The maximal number of cubes in a trivial Morse set for which an attempt
/// is made to join this set with another near-by Morse set.
/// Set to zero to suppress joining.
const int maxJoinSize = 2000;

/// The maximal size of a connecting orbit between two Morse sets
/// which can be considered for joining.
const int maxJoinConnection = 1000;

/// The maximal allowed distance between two Morse sets which can be
/// considered for joining. Note: It doesn't work with space wrapping.
const int maxJoinDistance = 10;


// --------------------------------------------------
// ----------------- other settings -----------------
// --------------------------------------------------

/// Should the ordering between the Morse sets be taken into consideration
/// while determining whether two Morse decompositions
/// computed for adjacent parameter boxes are in the same class?
const bool compareMorseOrdering = false;

/// Ignoring the isolation problem while matching Morse decompositions.
/// If this constant is set to "true" then Morse sets are matched
/// with each other indepent of whether the Conley index could be computed
/// for them or not. Otherwise, 'correct' and 'wrong' Morse sets
/// are matched with their counterparts in the same category only.
const bool ignoreIsolationForContinuation = true;

/// Ignoring the isolation problem while computing the Conley index.
/// If the constant is set to "true" then the index pair constructed on the
/// basis of a combinatorial Morse set with respect to the dynamics
/// restricted to the rectangular area of interest is also restricted
/// to that area by means of projecting the any boxes which stick out
/// of the region onto a single layer of boxes around the boundary.
const bool ignoreIsolationForConleyIndex = false;


} // namespace InfModel
} // namespace custom


#endif // _CMGRAPHS_P_H_

